/*   
    copyright: Steve Dekorte, 2004. All rights reserved.
    license: See _BSDLicense.txt.

    object: SkipDBRecord
    module: SkipDB
    description: 
*/

#include "SkipDBRecord.h"
#include "SkipDB.h"

SkipDBRecord *SkipDBRecord_new(void)
{
    SkipDBRecord *self = (SkipDBRecord *)calloc(1, sizeof(SkipDBRecord));
    self->pids = calloc(1, SKIPDB_MAX_LEVEL * sizeof(uint32_t));
    self->records = calloc(1, SKIPDB_MAX_LEVEL * sizeof(SkipDBRecord *));
    self->key = ByteArray_new();
    self->valuePid = 0;
    self->retainCount = 1;
    return self;
}

SkipDBRecord *SkipDBRecord_newWithDB_(void *db)
{
    SkipDBRecord *self = SkipDBRecord_new();
    SkipDBRecord_db_(self, db);
    return self;
}

void SkipDBRecord_free(SkipDBRecord *self)
{
    //printf("SkipDBRecord_free(%p)\n", (void *)self);
    if (self->records) 
    {
	int i;
	for (i = 0; i < self->level; i ++)
	{
	    SkipDBRecord *r = self->records[i];
	    if (r == self)
	    {
		printf("SkipDBRecord error: r == self\n");
		exit(1);
	    }
	    if (r) SkipDBRecord_release(r);
	}
	free(self->records);
    }
    if (self->pids) free(self->pids);
    ByteArray_free(self->key);
    self->retainCount --;
    free(self);
}

SkipDBRecord *SkipDBRecord_retain(SkipDBRecord *self)
{
    self->retainCount ++;
    return self;
}

void SkipDBRecord_release(SkipDBRecord *self)
{
    //printf("SkipDBRecord_release(%p)\n", (void *)self);
    self->retainCount --;
    if (self->retainCount < 1)
    {
	if (self->retainCount == -1)
	{
	    printf("SkipDBRecord ERROR: negative retain count\n");
	    exit(1);
	}
	SkipDBRecord_free(self);
    }
}

void SkipDBRecord_db_(SkipDBRecord *self, void *db)
{
    self->db = db;
}

void SkipDBRecord_pid_(SkipDBRecord *self, PID_TYPE pid)
{
    self->pid = pid;
}

PID_TYPE SkipDBRecord_pid(SkipDBRecord *self)
{
    return self->pid;
}

/* --------------------------------------- */

void SkipDBRecord_keyDatum_(SkipDBRecord *self, UDBDatum k)
{
    ByteArray_setData_length_(self->key, k.data, k.size);
}

ByteArray *SkipDBRecord_key(SkipDBRecord *self)
{
    return self->key;
}

UDB *SkipDBRecord_udb(SkipDBRecord *self)
{
    return ((SkipDB *)(self->db))->udb;
}

void SkipDBRecord_valueDatum_(SkipDBRecord *self, UDBDatum v)
{
    UDB *udb = SkipDBRecord_udb(self);
    if (self->valuePid)
    {
	UDB_at_put_(udb, self->valuePid, v);
    }
    else
    {
	self->valuePid = UDB_put_(udb, v);
	self->needsSave = 1;
    }
}

UDBDatum SkipDBRecord_valueDatum(SkipDBRecord *self)
{
    UDB *udb = SkipDBRecord_udb(self);
    return UDB_at_(udb, self->valuePid);
}

/* --- pointers ------------------------------------ */

int SkipDBRecord_level(SkipDBRecord *self)
{
    return self->level;
}

void SkipDBRecord_level_(SkipDBRecord *self, int level)
{
    self->level = level;
    self->pids = realloc(self->pids, level * sizeof(uint32_t));
    self->records = realloc(self->records, level * sizeof(SkipDBRecord *));
}

inline void SkipDBRecord_checkLevel_(SkipDBRecord *self, int level)
{
     if (level > self->level - 1) 
     { 
	printf("SkipDBRecord level out of range\n"); 
	exit(1); 
     }
}

void SkipDBRecord_atLevel_setPid_(SkipDBRecord *self, int level, PID_TYPE pid)
{
    SkipDBRecord_checkLevel_(self, level);
    self->pids[level] = pid;
    {
    SkipDBRecord *r = self->records[level];
    if (r) 
    {
	SkipDBRecord_release(r);
    }
    self->records[level] = 0x0;
    }
}

PID_TYPE SkipDBRecord_pidAtLevel_(SkipDBRecord *self, int level)
{
    SkipDBRecord_checkLevel_(self, level);
    return self->pids[level];
}

void SkipDBRecord_atLevel_setRecord_(SkipDBRecord *self, int level, SkipDBRecord *r)
{
    SkipDBRecord_checkLevel_(self, level);
    SkipDBRecord_atLevel_setPid_(self, level, SkipDBRecord_pid(r)); 
    self->records[level] = SkipDBRecord_retain(r);
}

SkipDBRecord *SkipDBRecord_cachedRecordAtLevel_(SkipDBRecord *self, int level)
{
    SkipDBRecord_checkLevel_(self, level);
    return self->records[level];
}

SkipDBRecord *SkipDBRecord_recordAtLevel_(SkipDBRecord *self, int level)
{
    SkipDBRecord_checkLevel_(self, level);
    {
	SkipDBRecord *r = self->records[level];
	if (!r)
	{
	    r = SkipDB_recordAtPid_(self->db, self->pids[level]);
	    self->records[level] = r; /* already retained */
	}
	return r;
    }
}

/* --- serialization ------------------------------------ */

void SkipDBRecord_toStream_(SkipDBRecord *self, BStream *s)
{
    BStream_writeTaggedByteArray_(s, self->key);
    BStream_writeInt32_(s, self->valuePid);
    {
	int i, max = SkipDBRecord_level(self);
	BStream_writeTaggedInt32_(s, max);
	for (i = 0; i < max; i ++)
	{
	    BStream_writeTaggedInt32_(s, SkipDBRecord_pidAtLevel_(self, i));
	}
    }
    //BStream_show(s);
}

void SkipDBRecord_fromStream_(SkipDBRecord *self, BStream *s)
{
    BStream_readTaggedByteArray_(s, self->key);
    self->valuePid = BStream_readInt32(s);
    {
	int i, max = BStream_readTaggedInt32(s);
	SkipDBRecord_level_(self, max);
	for (i = 0; i < max; i ++)
	{
	    SkipDBRecord_atLevel_setPid_(self, i, BStream_readTaggedInt32(s));
	}
    }
    //BStream_show(s);
}

/* --- search ------------------------------------ */

#define MIN(a, b) a < b ? a : b

int SkipDBRecord_compareKey_(SkipDBRecord *self, UDBDatum key)
{
    int s1 = ByteArray_length(self->key);
    int s2 = key.size;
    int min = MIN(s1, s2);
    int c = memcmp(ByteArray_bytes(self->key), key.data, min);
    
    if (s1 != s2 && c == 0)
    {
	return s1 > s2 ? 1 : -1;
    }
    
    return c;
}

SkipDBRecord *SkipDBRecord_find_(SkipDBRecord *self, UDBDatum key)
{
    SkipDB *db = self->db;
    int level = self->level;

    //printf("key: %s find: %s\n", ByteArray_asCString(self->key), key.data);
    //printf("find: %s in: ", key.data);
    //SkipDBRecord_show(self);

    while (level --)
    {
	SkipDBRecord *r = SkipDBRecord_recordAtLevel_(self, level);

	SkipDB_updateAt_put_(db, level, self);
		
	if (r)
	{
	    int c = SkipDBRecord_compareKey_(r, key);
	    
	    //printf("key: %s at level: %i\n", ByteArray_asCString(r->key), level);
	    if (ByteArray_length(r->key) == 0)
	    {
		printf("ERROR: key length = 0\n");
		exit(1);
	    }
	    	    
	    if (c == 0) /* record is a match */
	    { 
		return r; 
	    }
	    else if (c < 0) /* record key is smaller */
	    {
		return SkipDBRecord_find_(r, key);
	    }
	    /* otherwise record key is bigger so we keep searching */
	}
    }
    return 0x0;
}

void SkipDBRecord_copyLevel_from_(SkipDBRecord *self, int level, SkipDBRecord *other)
{
    PID_TYPE newPid = SkipDBRecord_pidAtLevel_(other, level);

    SkipDBRecord *cachedRecord = SkipDBRecord_cachedRecordAtLevel_(other, level);
    if (cachedRecord)
    {
	SkipDBRecord_atLevel_setRecord_(self, level, cachedRecord);
    }
    else
    {
	SkipDBRecord_atLevel_setPid_(self, level, newPid);
    }
}

void SkipDBRecord_willRemove_(SkipDBRecord *self, SkipDBRecord *other)
{
    PID_TYPE otherPid = SkipDBRecord_pid(other);
    int level = self->level;
    int needsSave = 0;
    
    while (level --)
    {
	if (SkipDBRecord_pidAtLevel_(self, level) == otherPid)
	{
	    SkipDBRecord_copyLevel_from_(self, level, other);
	    needsSave = 1;
	}
    }
    
    if (needsSave) SkipDB_updateRecord_(self->db, self);
}

void SkipDBRecord_show(SkipDBRecord *self)
{
    int i;

    //printf("SRec %i '%s' (", self->pid, ByteArray_asCString(self->key));
    //printf("SRec '%s' (", ByteArray_asCString(self->key));
	
    for (i = 0; i < self->level; i ++)
    {
	if ( self->records[i])
	{
	    printf("'%s'", ByteArray_asCString(self->records[i]->key));
	}
	else
	{
	    printf("%i", self->pids[i]);
	}
	
	if (i != self->level - 1) printf(", ");
    }
    //printf("  retainCount: %i\n", self->retainCount);
    printf(")\n");
}
