/* Scheduler
 * Based on Edgar Toernig's Minimalistic cooperative multitasking
 * http://www.goron.de/~froese/
 * reorg by Steve Dekorte and Chis Double
 * Symbian and Cygwin support by Chis Double
 * Linux/PCC, Linux/Opteron, Irix and FreeBSD/Alpha, ucontext support by Austin Kurahone
 * FreeBSD/Intel support by Faried Nawaz
 * Mingw support by Pit Capitain
 * Visual C support by Daniel Vollmer
 * Solaris support by Manpreet Singh
 */

#include "Scheduler.h" 
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "PortableGettimeofday.h"
#include "PortableUsleep.h"
#include <stddef.h>

#if defined(HAS_UCONTEXT)
#include <ucontext.h>
#endif

/* since currentScheduler is static, this code isn't safe for 
 * use in a scheduler-per-vm model unless you
 * add a bit of code for your platform to set it properly or
 * link it to some coro dependent variable.
 */

#if !defined(__SYMBIAN32__)
static Scheduler *currentScheduler; 
#endif

#if defined(__SYMBIAN32)
void ProcessUIEvent();
#endif

 /*#define SCHEDULER_DEBUG 1 */

/* This is the system dependent function.
 * Setup a jmp_buf so when we longjmp, it will invoke 'func' using 'stack'.
 * Important: 'func' must not return!
 *
 * Usually done by setting the program counter and stack pointer of a new, empty stack.
 * If you're adding a new platform, look in the setjmp.h for PC and SP members 
 * of the stack structure
 *
 * If you don't see those members, Kentaro suggests writting a simple 
 * test app that calls setjmp and dumps out the contents of the jmp_buf.  
 * (The PC and SP should be in jmp_buf->__jmpbuf).  
 *
 * Using something like GDB to be able to peek into register contents right 
 * before the setjmp occurs would be helpful also.
 */

static void setup_jmpbuf(jmp_buf buf, void *func, void *stack, size_t stacksize)
{
  /* Solaris 9 Sparc with GCC */
  #if defined(__SVR4) && defined (__sun)
    #if defined(_JBLEN) && (_JBLEN == 12) && defined(__sparc)
      #if defined(_LP64) || defined(_I32LPx)
        #define JBTYPE long
        JBTYPE x;
      #else
        #define JBTYPE int
        JBTYPE x;
        asm("ta 3"); /* flush register window */
      #endif
      #define SUN_PROGRAM_COUNTER 2
      #define SUN_STACK_START_INDEX 3 
      #define SUN_STACK_END_INDEX 1
    /* Solaris 9 i386 with GCC */
    #elif defined(_JBLEN) && (_JBLEN == 10) && defined(__i386)
      #if defined(_LP64) || defined(_I32LPx)
        #define JBTYPE long
        JBTYPE x;
      #else
        #define JBTYPE int
        JBTYPE x;
      #endif
      #define SUN_PROGRAM_COUNTER 5
      #define SUN_STACK_START_INDEX 3 
      #define SUN_STACK_END_INDEX 4
    #endif
  #endif
  /*
  #ifdef SCHEDULER_DEBUG
  printf("Scheduler setup_jmpbuf\n");
  #endif
  */
  setjmp(buf);
  
  /* #if defined(__MACOSX__) && defined(_BSD_PPC_SETJMP_H_) */
  #if defined(_BSD_PPC_SETJMP_H_) 
    /* OSX/PPC */
    #define setjmp _setjmp /* no need for signal masks */
    #define longjmp _longjmp
    memset(((char *)stack) + stacksize - 64, 0, 64);	/* very crude... */
    buf[0] = ((intptr_t)stack + stacksize - 64 + 15) & ~15;
    buf[21] = (intptr_t)func;

  #elif defined(sgi) && defined(_IRIX4_SIGJBLEN)
    /* Irix/SGI */
    buf[JB_SP] = (__uint64_t)((char *)stack + stacksize - 8);
    buf[JB_PC] = (__uint64_t)func;

  #elif defined(linux)
    /* Various flavors of Linux. */
    #if defined(JB_GPR1)
      /* Linux/PPC */
      buf->__jmpbuf[JB_GPR1] = ((int) stack + stacksize - 64 + 15) & ~15;
      buf->__jmpbuf[JB_LR] = (int) func;
    
    #elif defined(JB_RBX)
      /* Linux/Opteron */
      buf->__jmpbuf[JB_RSP] = (long int )stack + stacksize;
      buf->__jmpbuf[JB_PC] = func;

    #elif defined(JB_SP)
      /* Linux/x86 with glibc2 */
      buf->__jmpbuf[JB_SP] = (int)stack + stacksize;
      buf->__jmpbuf[JB_PC] = (int)func;
    
    #elif defined(_I386_JMP_BUF_H)
      /* x86-linux with libc5 */
      buf->__sp = stackend;
      buf->__pc = func;
    
    #elif defined(__JMP_BUF_SP)
      /* arm-linux on the sharp zauras */
      buf->__jmpbuf[__JMP_BUF_SP] = (int)stack + stacksize;
      buf->__jmpbuf[__JMP_BUF_SP+1] = (int)func;

    #else
      goto error;
    #endif

  #elif defined(__CYGWIN__) 
    /* Cygwin */
    buf[7] = (int)((int)stack + stacksize);
    buf[8] = (int)func;
    
  #elif defined(__MINGW32__) 
    buf[4] = (int)((unsigned char *)stack + stacksize - 100);   /* esp */
    buf[5] = (int)func; /* eip */
      
  #elif defined(_MSC_VER)
    /* win32 visual c */
    /* should this be the same as __MINGW32__? */
    buf[4] = (int)((unsigned char *)stack + stacksize);  /* esp */
    buf[5] = (int)func; /* eip */
    
  #elif defined(__SYMBIAN32__)
    /* Symbinan. */
    buf[0] = 0;
    buf[1] = 0;
    buf[2] = 0;
    buf[3] = (unsigned int)stack + CSTACK_STACKSIZE - 64;
    buf[9] = (int)func;
    buf[8] =  buf[3] + 32;

  #elif defined(__FreeBSD__)
    /* FreeBSD. */
    #if defined(_JBLEN) && (_JBLEN == 81)
      /* FreeBSD/Alpha */
      buf->_jb[2] = (long)func;     /* sc_pc */
      buf->_jb[26+4] = (long)func;  /* sc_regs[R_RA] */
      buf->_jb[27+4] = (long)func;  /* sc_regs[R_T12] */
      buf->_jb[30+4] = (long)(stack + stacksize); /* sc_regs[R_SP] */

    #elif defined(_JBLEN)
      /* FreeBSD on IA32 */
      buf->_jb[2] = (long)(stack + stacksize);
      buf->_jb[0] = (long)func;

    #else
      goto error;
    #endif

  #elif defined(__NetBSD__)
    /* NetBSD. */
    #if defined(_JB_ATTRIBUTES)
      /* NetBSD i386. */
      buf[2] = (long)(stack + stacksize);
      buf[0] = (long)func;

    #else
      goto error;

    #endif

  #elif defined(__SVR4) && defined(__sun)
    /* Solaris. */
    #if defined(SUN_PROGRAM_COUNTER)
      /* SunOS 9 */
      buf[SUN_PROGRAM_COUNTER] = (JBTYPE)func;
    
      x = (JBTYPE)stack;
      while ((x % 8) != 0) x--; /* align on an even boundary */
      buf[SUN_STACK_START_INDEX] = (JBTYPE)x;
      x = (JBTYPE)((JBTYPE)stack-stacksize/2+15);
      while ((x % 8) != 0) x++; /* align on an even boundary */
      buf[SUN_STACK_END_INDEX] = (JBTYPE)x;

    #else
      goto error;
    #endif

  #else
    goto error;
  #endif

  return;

    if (0) { goto error; } /* Shut GCC up in the good case. */
error:
    printf("Io Scheduler error: no setup_jmpbuf entry for this platform\n.");
    exit(1);
}

#if defined(HAS_UCONTEXT)
static void setup_ucontext(Coro *self, void *func, void *stack, size_t stacksize)
{
    ucontext_t *ucp = (ucontext_t *) &self->env;

    getcontext(ucp);
    ucp->uc_stack.ss_sp = stack;
    ucp->uc_stack.ss_size = stacksize;
    ucp->uc_stack.ss_flags = 0;
    ucp->uc_link = NULL;
    makecontext(ucp, func, 0);
}
#endif

#if !defined(__SYMBIAN32__)
void Scheduler_current_(Scheduler *self)
{
    currentScheduler = self;
}
#endif

Scheduler *Scheduler_new(void *state)
{
    Scheduler *self = (Scheduler *)calloc(1, sizeof(Scheduler));
    self->state = state;
    self->pausedCoros = List_new();
    self->recycledCoros = List_new();
    self->mainCoro = Coro_new();
    Coro_label_(self->mainCoro, "mainCoro");
    Coro_scheduler_(self->mainCoro, self);
    self->currentCoro = self->mainCoro;
    self->coroCount = 1;
    self->context = 0x0;
    self->start_func = 0x0;
    self->sleepMicroseconds = 0; 
    self->maxToRecycle = 10000;

    Scheduler_current_(self);

    return self;
}

void Scheduler_free(Scheduler *self)
{
    Coro *coro;
    
    while (self->currentCoro != self->mainCoro)
    {
	printf("Scheduler error: attempt to free outside of main coroutine\n");
	exit(1);
    }
    
    coro = (Coro *)List_first(self->pausedCoros);
    if (coro) Coro_freeCoroChain(coro);
    List_free(self->pausedCoros);
    
    coro = (Coro *)List_first(self->recycledCoros);
    
    if (coro) 
    {
	Coro_freeCoroChain(coro);
    }
    
    List_free(self->recycledCoros);
    Coro_freeCoroChain(self->currentCoro);
    free(self);
}

Coro *Scheduler_mainCoro(Scheduler *self) 
{ 
    return self->mainCoro; 
}

Coro *Scheduler_currentCoro(Scheduler *self) 
{ 
    return self->currentCoro; 
}

void Scheduler_secondsInSleepPeriod_(Scheduler *self, float seconds)
{ 
    self->sleepMicroseconds = (int)(seconds * 1000000); 
    //printf("Scheduler_secondsInSleepPeriod_ usleep(%i)\n", (int)self->sleepMicroseconds);
}

float Scheduler_secondsInSleepPeriod(Scheduler *self)
{ 
    return ((float)(self->sleepMicroseconds)) / 1000000.0; 
}

unsigned int Scheduler_coroCount(Scheduler *self) 
{ 
    return self->coroCount; 
}

unsigned int Scheduler_hasCoro_(Scheduler *self, Coro *coro)
{ 
    return Coro_hasCoro_(self->currentCoro, coro); 
}

#if !defined(__SYMBIAN32__)
Scheduler *Scheduler_current(void) 
{ 
    return currentScheduler; 
}
#endif

static void start_coro(void)
{
    Scheduler *self = Scheduler_current();
    self->start_func(self->context);
    printf("Scheduler error: returned from coro start function\n");
    exit(1);
    Scheduler_killCoro_(self, self->currentCoro);
}

Coro *Scheduler_newCoro(Scheduler *self)
{
    Coro *coro = (Coro *)List_pop(self->recycledCoros);
    
    if (!coro) 
    {
	coro = Coro_new();
	Coro_scheduler_(coro, self);
    }
    
    return coro;    
}

/*
 Coro *Scheduler_createCoro(Scheduler *self, void *aContext, void (*func)(void *))
 {
     Coro *newCoro = Scheduler_newCoro(self);
     Scheduler_startCoro(self, coro, aContext, func);
     return newCoro;
 }
 */

void Scheduler_startCoro(Scheduler *self, Coro *coro, void *aContext, void (*func)(void *))
{
#ifdef SCHEDULER_DEBUG
    printf("Scheduler create coro %p\n", (void *)coro);
    printf("self->currentCoro = %p %s\n", (void *)self->currentCoro, Coro_label(self->currentCoro));
#endif

    Scheduler_current_(self);
#if defined(HAS_UCONTEXT)
    setup_ucontext(coro, (void *) start_coro, coro + 1, CSTACK_STACKSIZE);
#else    
    /* setjmp/longjmp is flakey under Symbian. If the setjmp is done
	inside the setup_jmpbuf call then a crash occurs. inlining it
	here solves the problem 
	*/
#if !defined(__SYMBIAN32__)
    setup_jmpbuf(coro->env, (void *)start_coro, coro + 1, CSTACK_STACKSIZE);
#else
    CSTACK_SETJMP(coro->env);
    coro->env[0] = 0;
    coro->env[1] = 0;
    coro->env[2] = 0;
    coro->env[3] = (unsigned int)(newCoro + 1) + CSTACK_STACKSIZE - 64;
    coro->env[9] = (int)start_coro;
    coro->env[8] =  coro->env[3] + 32;
#endif

#endif
    
    Coro_insertBefore_(coro, self->currentCoro);
    
    self->coroCount++;
    Scheduler_sendUpdateCallback(self);
    
    self->start_func = func;
    self->context = aContext;

#if defined(HAS_UCONTEXT)
    self->currentCoro = self->currentCoro->prev;
    swapcontext(&self->currentCoro->next->env, &self->currentCoro->env);
#else
    if (CSTACK_SETJMP(self->currentCoro->env) == 0)
    {
	self->currentCoro = self->currentCoro->prev;
	CSTACK_LONGJMP(*(jmp_buf*)self->currentCoro->env, 1);
    }
#endif
}


inline void Scheduler_yieldSleepInline(Scheduler *self)
{
    /* on BSD, nanosleep is a busy wait unless another process is active */
    /*
#ifdef _TIMESPEC_DECLARED
    if (self->nanoSleepPeriod)
    {
	struct timespec rqtp;  
	struct timespec extra;
	rqtp.tv_sec = 0;
	rqtp.tv_nsec = self->nanoSleepPeriod; 
	nanosleep(&rqtp, &extra);

	// if unmasked signal is raised, nanosleep returns -1 
	//while (nanosleep(&rqtp, &extra) == -1)
	//{
	//    rqtp.tv_sec  = extra.tv_sec;
	//    rqtp.tv_nsec = extra.tv_nsec;     
	//}
    }
#endif
    */
    if (self->sleepMicroseconds)
    {
	unsigned int ms = self->sleepMicroseconds / self->coroCount;
	/*printf("Scheduler usleep(%i)\n", (int)ms);*/
	usleep(ms); 
    }
}

void Scheduler_yield(Scheduler *self)
{
#ifdef SCHEDULER_DEBUG
    printf("Scheduler yield %s_%p\n", Coro_label(self->currentCoro), (void *)self->currentCoro);
#endif
    
    if (self->currentCoro == self->currentCoro->next) 
    { 
	return; 
    }
    
    Scheduler_yieldSleepInline(self);
    
#if defined(__SYMBIAN32)
    ProcessUIEvent();
#endif
    /*
#ifdef SCHEDULER_DEBUG
     printf("Scheduler setjmp   %p %s\n", (void *)self->currentCoro, Coro_label(self->currentCoro));
#endif
     */
#if defined(HAS_UCONTEXT)
    if (1) 
    {
        Coro *prevCoro = self->currentCoro;
#else
    if (setjmp(self->currentCoro->env) == 0)
    {
#endif
	Coro_check(self->currentCoro);
	
	self->currentCoro = self->currentCoro->next;
	
	while (Coro_shouldPause(self->currentCoro) || Coro_shouldKill(self->currentCoro))
	{
	    Coro *next = self->currentCoro->next;
	    
	    if (Coro_shouldPause(self->currentCoro)) 
	    { 
		Scheduler_pauseCoro_(self, self->currentCoro); 
	    }
	    else if (Coro_shouldKill(self->currentCoro)) 
	    { 
		Scheduler_killCoro_(self, self->currentCoro); 
	    }
	    
	    self->currentCoro = next;
	    Coro_check(self->currentCoro);
	} 
	
#ifdef SCHEDULER_DEBUG
	printf("Scheduler longjmp  %p %s\n", 
	       (void *)self->currentCoro, Coro_label(self->currentCoro));
	
	Scheduler_show(self);
#endif
	
	Coro_doUnyieldCallback(self->currentCoro);
    
#if defined(HAS_UCONTEXT)
        swapcontext(&prevCoro->env, &self->currentCoro->env);
#else
	longjmp(self->currentCoro->env, 1);
#endif
    }
}

void Scheduler_show(Scheduler *self)
{
    Coro *coro = self->mainCoro;
    
    for (;;) 
    {
	Coro_show(coro);
	coro = coro->next;
	if (coro == self->mainCoro) break;
	printf(" -> ");
    }
    printf("\n");
}

/* --- kill coro --------------------------------------------- */

void Scheduler_prepareToKillCoro_(Scheduler *self, Coro *coro)
{
#ifdef SCHEDULER_DEBUG
    printf("Scheduler prepare kill %s_%p\n", Coro_label(coro), (void *)coro);
#endif
    
    Coro_shouldKill_(coro, 1);
    /*if (self->currentCoro == coro) Scheduler_yield(self); */
}

void Scheduler_killCoro_(Scheduler *self, Coro *coro)
{
#ifdef SCHEDULER_DEBUG
    printf("Scheduler kill %s_%p\n", Coro_label(coro), (void *)self->currentCoro);
#endif
    
    List_append_(self->recycledCoros, coro);
    Coro_remove(coro);
    self->coroCount --;
    Scheduler_sendUpdateCallback(self);
    
    /* free excess recycled Coros */
    while (List_size(self->recycledCoros) > self->maxToRecycle)
    { 
	Coro_free((Coro *)List_pop(self->recycledCoros)); 
    }
}

/* --- pause/resume coro --------------------------------------------- */

Coro *Scheduler_prepareToPauseCurrentCoro(Scheduler *self)
{
    Scheduler_prepareToPauseCoro_(self, self->currentCoro);
    return self->currentCoro;
}

void Scheduler_prepareToPauseCoro_(Scheduler *self, Coro *coro)
{
#ifdef SCHEDULER_DEBUG
    printf("Scheduler prepare pause %s_%p\n", Coro_label(coro), (void *)coro);
#endif
    Coro_shouldPause_(coro, 1);
    /*if (self->currentCoro == coro) Scheduler_yield(self); */
}

void Scheduler_pauseCoro_(Scheduler *self, Coro *coro)
{
    Coro_shouldPause_(coro, 0);
    Coro_remove(coro);
    self->coroCount --;
    Scheduler_sendUpdateCallback(self);
#ifdef SCHEDULER_DEBUG
    printf("Scheduler paused coro %s_%p\n", Coro_label(coro), (void *)coro);
#endif
    List_append_(self->pausedCoros, coro);
}

int Scheduler_hasPausedCoro_(Scheduler *self, Coro *coro)
{
    return List_contains_(self->pausedCoros, coro);
}

void Scheduler_resumeCoro_(Scheduler *self, Coro *coro)
{
    /* don't resume if... */
    if (Scheduler_hasCoro_(self, coro) || /* it's already active... */
	!List_contains_(self->pausedCoros, coro) /* it's not paused... */
	) 
    {
	if (Coro_shouldPause(coro))
	{
#ifdef SCHEDULER_DEBUG
	    printf("Scheduler resuming not yet paused coro %s\n", Coro_label(coro));
#endif
	    Coro_shouldPause_(coro, 0);
	}
	else
	{
#ifdef SCHEDULER_DEBUG
	    if (Scheduler_hasCoro_(self, coro))
	    {
		if (Coro_shouldPause(coro))
		{
		    Coro_shouldPause_(coro, 0);
		}
		else
		{
		    printf("Scheduler CANNOT resume Coro_%p %s, already running ------\n", 
		       (void *)coro, Coro_label(coro));
		
		    Scheduler_show(self);
		    Scheduler_hasCoro_(self, coro);
		}
	    }
	    else
	    {
		printf("Scheduler CANNOT resume %s, not in paused list --------\n", 
		       Coro_label(coro));
	    }      
#endif
	}    
	return;
    }

    #ifdef SCHEDULER_DEBUG
    printf("Scheduler resume coro %p %s\n", (void *)coro, Coro_label(coro));
    #endif

    Coro_insertAfter_(coro, self->currentCoro);
    self->coroCount ++;
    List_remove_(self->pausedCoros, coro);
    Coro_shouldPause_(coro, 0);
}

/*
 void Scheduler_pause(Scheduler *self)
 { 
     Scheduler_prepareToPauseCoro_(self, self->currentCoro); 
 }
 
 void Scheduler_resume(Scheduler *self)
 { 
     Scheduler_resumeCoro_(self, self->currentCoro); 
 }
 */

/* --- wait --------------------------------------------- */

//extern double IoState_endTime(struct IoState *self);

double Scheduler_waitSeconds_(Scheduler *self, double seconds)
{
    /* 
    * This implementation is simple but inefficient.
    * Pausing the waiting coros and having a separate timer stack coro
    * would be much faster when multiple coros are waiting.
    */    
    struct timeval startTv, tmp;
    double start, now, end, stateEnd = 0; //IoState_endTime((struct IoState *)(self->state));
    
    /* Verify the time to wait is not negative */
    if (seconds < 0) 
    {
	seconds = 0;
    }
    
    /* Get the start and current time */
    gettimeofday(&startTv, NULL);
    start = (double)startTv.tv_sec + ((double)startTv.tv_usec/1000000.0);
    now = start;
    
    /* Calculate the end of time */
    end = now + seconds;
    
    /* Wait for time to pass */
    while (end > now && (stateEnd <= 0 || stateEnd > now)) 
    {
	Scheduler_yield(self);
	gettimeofday(&tmp, NULL);
	now = (double)tmp.tv_sec + ((double)tmp.tv_usec/1000000.0);
    }
    
    /* return the number of seconds actually waited */
    return (now - start);
}

/* --- callbacks --------------------------------------------- */

void Scheduler_callbackContext_(Scheduler *self, void *context)
{
    self->callbackContext = context;
}

void Scheduler_updateCallback_(Scheduler *self, SchedulerUpdateCallback *func)
{
    self->updateCallback = func;
}

void Scheduler_sendUpdateCallback(Scheduler *self)
{
    if (self->updateCallback) 
    {
	self->updateCallback(self->callbackContext, self->coroCount);
    }
}
